// Fractal generator node

// Written by Bernie Roehl, November 1996

// Inspired by Steve Anger's FRGEN

package fractal;

import vrml.*;
import vrml.field.*;
import vrml.node.*;
import java.io.*;
import java.util.*;
import utility.*;

public class TriFractal extends Script {
    Vec3 nbias;      // noise bias vector
    Vec3 nscale;     // noise scale vector
    Vec3 nfalloff;   // noise falloff vector
    int depth;       // recursion depth
    int seed;        // random number seed

    Random rand = new Random();

    Vector outputVertices = new Vector();   // for Coordinate node
    Vector outputTriangles = new Vector();  // for coordIndex field

    public void initialize() {
        depth = ((SFInt32) getField("recursionDepth")).getValue();
        seed = ((SFInt32) getField("randomSeed")).getValue();
        nbias = convertVector((SFVec3f) getField("bias"));
        nscale = convertVector((SFVec3f) getField("scale"));
        nfalloff = convertVector((SFVec3f) getField("falloff"));

        if (seed != 0)
            rand.setSeed(seed);

        int nvertices = ((MFVec3f) getField("triangles")).getSize();
        int ntriangles = nvertices / 3;

        float[][] vertices = new float[nvertices][3];
        ((MFVec3f) getField("triangles")).getValue(vertices);

        int[] fixFlags = new int[nvertices];
        ((MFInt32) getField("fixFlags")).getValue(fixFlags);

        for (int i = 0; i < ntriangles; ++i) {
            Vec3 a = new Vec3(vertices[i*3][0], vertices[i*3][1], vertices[i*3][2]);
            Vec3 b = new Vec3(vertices[i*3+1][0], vertices[i*3+1][1], vertices[i*3+1][2]);
            Vec3 c = new Vec3(vertices[i*3+2][0], vertices[i*3+2][1], vertices[i*3+2][2]);
            boolean fix_ab = (fixFlags[i*3+0] != 0);
            boolean fix_bc = (fixFlags[i*3+1] != 0);
            boolean fix_ca = (fixFlags[i*3+2] != 0);
            tri_fractal(a, b, c, fix_ab, fix_bc, fix_ca, depth);
        }  

        StringBuffer coordBuff = new StringBuffer(1024);
        coordBuff.append("Coordinate { point [\n");
        Enumeration en = outputVertices.elements();
        while (en.hasMoreElements()) {
            Vec3 v = (Vec3) en.nextElement();
            coordBuff.append("\t" + v + "\n");
        }
        coordBuff.append(" ] }\n");

        try {
            SFNode coord_changed = (SFNode) getEventOut("coord_changed");
            coord_changed.setValue(getBrowser().createVrmlFromString(coordBuff.toString())[0]);
        }
        catch (InvalidVRMLSyntaxException ex) {}

        int[] points = new int[outputTriangles.size() * 4];
        en = outputTriangles.elements();
        for (int n = 0; en.hasMoreElements(); n += 4) {
            Triangle tri = (Triangle) en.nextElement();
            // note that we reverse the order to counter-clockwise
            points[n] = tri.getC();
            points[n+1] = tri.getB();
            points[n+2] = tri.getA();
            points[n+3] = -1;  // end the face
        }

        MFInt32 coordIndex_changed = (MFInt32) getEventOut("coordIndex_changed");
        coordIndex_changed.setValue(points);

    }

    protected void tri_fractal(Vec3 a, Vec3 b, Vec3 c,
                            boolean fix_ab, boolean fix_bc, boolean fix_ca,
                            int level) {
        if (level <= 0) {  // end of the recursion; dump a triangle
            addTriangle(a, b, c);
            return;
        }

        // find midpoints:
        Vec3 ab = new Vec3((a.add(b)).scale(0.5f));
        Vec3 bc = new Vec3((b.add(c)).scale(0.5f));
        Vec3 ca = new Vec3((c.add(a)).scale(0.5f));

        // find lengths of sides:
        float ab_len = (a.sub(b)).mag();
        float bc_len = (b.sub(c)).mag();
        float ca_len = (c.sub(a)).mag();

        // find normal to the triangle:
        Vec3 norm = (b.sub(a)).cross(c.sub(a)).normalize();

        // compute noise vectors
        Vec3 noise_ab = noise_vector(ab, depth - level).scale(ab_len);
        Vec3 noise_bc = noise_vector(bc, depth - level).scale(bc_len);
        Vec3 noise_ca = noise_vector(ca, depth - level).scale(ca_len);

        // check for "fixed" edges
        if (fix_ab) noise_ab = noise_ab.cross(norm);
        if (fix_bc) noise_bc = noise_bc.cross(norm);
        if (fix_ca) noise_ca = noise_ca.cross(norm);

        // perturb the three vectors
        ab = ab.add(noise_ab);
        bc = bc.add(noise_bc);
        ca = ca.add(noise_ca);

        // now recursively process the sub-triangles
        tri_fractal(a, ab, ca, fix_ab, false, fix_ca, level-1);
        tri_fractal(b, bc, ab, fix_bc, false, fix_ab, level-1);
        tri_fractal(c, ca, bc, fix_ca, false, fix_bc, level-1);
        tri_fractal(ab, bc, ca, false, false, false, level-1);
    }

    protected Vec3 noise_vector(Vec3 v, int level) {
        rand.setSeed(seed   // random vector is based on v, for consistency
            ^ (int) (-23465 * v.getX())
            ^ (int)  (17365 * v.getY())
            ^ (int)   (5364 * v.getZ()));

        double xscale = nscale.getX() * Math.exp(level * Math.log(nfalloff.getX()));
        double yscale = nscale.getY() * Math.exp(level * Math.log(nfalloff.getY()));
        double zscale = nscale.getZ() * Math.exp(level * Math.log(nfalloff.getZ()));
        Vec3 randvec = new Vec3( 2 * rand.nextFloat() - 1, 2 * rand.nextFloat() - 1, 2 * rand.nextFloat() - 1);
        randvec = randvec.add(nbias);
        return new Vec3(
            (float) xscale * randvec.getX(),
            (float) yscale * randvec.getY(),
            (float) zscale * randvec.getZ()
            );
    }

    protected Vec3 convertVector(SFVec3f sfvec) {
        return new Vec3(sfvec.getX(), sfvec.getY(), sfvec.getZ());
    }

    protected void addTriangle(Vec3 a, Vec3 b, Vec3 c) {
        int a_index = -1, b_index = -1, c_index = -1;
        int i;
        Enumeration en = outputVertices.elements();
        for (i = 0; en.hasMoreElements(); ++i) {
            Vec3 v = (Vec3) en.nextElement();
            if (v.equals(a)) a_index = i;
            if (v.equals(b)) b_index = i;
            if (v.equals(c)) c_index = i;
        }
        if (a_index == -1) {
            a_index = i++;
            outputVertices.addElement(a);
        }
        if (b_index == -1) {
            b_index = i++;
            outputVertices.addElement(b);
        }
        if (c_index == -1) {
            c_index = i++;
            outputVertices.addElement(c);
        }
        outputTriangles.addElement(new Triangle(a_index, b_index, c_index));
    }

}

