// Class for handling 6 degree-of-freedom devices

// Written by Bernie Roehl, February 1997

public class Input6D {
    protected int port;        // set by constructor

    // read by poll():
    protected float rawX, rawY, rawZ;
    protected float rawR, rawU, rawV;
    protected int buttons;
    protected int whichButton;
    protected int status;
    protected boolean changed;
    protected boolean buttonChanged;

    // processed values:
    protected float x, y, z;
    protected float r, u, v;

    // set by setup():
    protected String description;  // name and description
    protected int nbuttons;        // number of buttons
    protected float minX, maxX, minY, maxY, minZ, maxZ;
    protected float minR, maxR, minU, maxU, minV, maxV;
    protected float centerX, centerY, centerZ;
    protected float centerR, centerU, centerV;
    protected boolean allsWell;

    // set by application:
    protected float xScale = 1f, yScale = 1f, zScale = 1f;
    protected float rScale = 1f, uScale = 1f, vScale = 1f;

    protected boolean accumulateTranslations = false;
    protected boolean accumulateRotations = false;

    protected long lastPolled;  // time

    public float getX() { return x; }
    public float getY() { return y; }
    public float getZ() { return z; }
    public float getR() { return r; }
    public float getU() { return u; }
    public float getV() { return v; }
    public int getButton() { return whichButton; }
    public int getButtons() { return buttons; }
    public int getStatus() { return status; }
    public boolean hasChanged() { return changed; }
    public boolean buttonHasChanged() { return buttonChanged; }

    public void setX(float value) { x = value; }
    public void setY(float value) { y = value; }
    public void setZ(float value) { z = value; }
    public void setR(float value) { r = value; }
    public void setU(float value) { u = value; }
    public void setV(float value) { v = value; }

    public String getDescription() { return description; }
    public int getNumberOfButtons() { return nbuttons; }
    public int getPort() { return port; }
    public void setPort(int p) { port = p; }

    public float getScaleX() { return xScale; }
    public float getScaleY() { return yScale; }
    public float getScaleZ() { return zScale; }
    public void setScaleX(float x) { xScale = x; }
    public void setScaleY(float y) { yScale = y; }
    public void setScaleZ(float z) { zScale = z; }

    public float getScaleR() { return rScale; }
    public float getScaleU() { return uScale; }
    public float getScaleV() { return vScale; }
    public void setScaleR(float r) { rScale = r; }
    public void setScaleU(float u) { uScale = u; }
    public void setScaleV(float v) { vScale = v; }

    public boolean getAccumulateTranslations() { return accumulateTranslations; }
    public boolean getAccumulateRotations() { return accumulateRotations; }
    public void setAccumulateTranslations(boolean value) { accumulateTranslations = value; }
    public void setAccumulateRotations(boolean value) { accumulateRotations = value; }

    public Input6D(int prt) {
        port = prt;
        setup();
    }

    private native void nativeSetup();

    private native void nativePoll();

    public void setup() {
        allsWell = true;
        nativeSetup();
        if (status != 0) {
            description = "Could not initialize; error code " + status; 
            allsWell = false;
        }
        // read the initial values, use them as the centerpoint
        changed = buttonChanged = false;
        nativePoll();
        lastPolled = System.currentTimeMillis();
        centerX = rawX;  centerY = rawY;  centerZ = rawZ;
        centerR = rawR;  centerU = rawU;  centerV = rawV;
        maxX -= centerX;  maxY -= centerY;  maxZ -= centerZ;
        maxR -= centerR;  maxU -= centerU;  maxV -= centerV;
        minX -= centerX;  minY -= centerY;  minZ -= centerZ;
        minR -= centerR;  minU -= centerU;  minV -= centerV;
        if (maxX == 0) maxX = 1;  if (minX == 0) minX = -1;
        if (maxY == 0) maxY = 1;  if (minY == 0) minY = -1;
        if (maxZ == 0) maxZ = 1;  if (minZ == 0) minZ = -1;
        if (maxR == 0) maxR = 1;  if (minR == 0) minR = -1;
        if (maxU == 0) maxU = 1;  if (minU == 0) minU = -1;
        if (maxV == 0) maxV = 1;  if (minV == 0) minV = -1;
    }

    public void poll() {
        // update the lastPolled time
        long now = System.currentTimeMillis();
        long deltaT = now - lastPolled;
        lastPolled = now;
        if (allsWell == false) {  // not initialized
            status = -1;
            return;
        }
        // poll the device
        changed = buttonChanged = false;
        nativePoll();
        if (!changed)
            return;
        // shift to center
        rawX -= centerX;  rawY -= centerY;  rawZ -= centerZ;
        rawR -= centerR;  rawU -= centerU;  rawV -= centerV;
        // scale and fit to range
        float dx = xScale * rawX / ((rawX < 0) ? -minX : maxX);
        float dy = yScale * rawY / ((rawY < 0) ? -minY : maxY);
        float dz = zScale * rawZ / ((rawZ < 0) ? -minZ : maxZ);
        float dr = rScale * rawR / ((rawR < 0) ? -minR : maxR);
        float du = uScale * rawU / ((rawU < 0) ? -minU : maxU);
        float dv = vScale * rawV / ((rawV < 0) ? -minV : maxV);
        // accumulate
        if (accumulateTranslations) {
            x += dx * deltaT / 1000;
            y += dy * deltaT / 1000;
            z += dz * deltaT / 1000;
        }
        else {
            x = dx;
            y = dy;
            z = dz;
        }
        if (accumulateRotations) {
            r += dr * deltaT / 1000;
            u += du * deltaT / 1000;
            v += dv * deltaT / 1000;
        }
        else {
            r = dr;
            u = du;
            v = dv;
        }
    }

    static { System.loadLibrary("Input6D"); }
}

