// Script for reading 6 degree-of-freedom input devices

// Written by Bernie Roehl, February 1997

import vrml.*;
import vrml.node.*;
import vrml.field.*;

public class Read6DOF extends Script {
    Input6D device;
    boolean enabled;
    boolean latching;
    SFVec3f position;
    SFRotation orientation;
    SFInt32 buttons;
    SFInt32 whichButton;
    SFBool[] buttonArray = new SFBool[8];

    public void initialize() {
        int port = ((SFInt32) getField("port")).getValue();
        device = new Input6D(port);

        // set accumulation flags
        device.setAccumulateTranslations(((SFBool) getField("accumulateTranslations")).getValue());
        device.setAccumulateRotations(((SFBool) getField("accumulateRotations")).getValue());

        // get translation scale
        SFVec3f vec = (SFVec3f) getField("translationScale");
        float[] scale = new float[3];
        vec.getValue(scale);
        device.setScaleX(scale[0]);
        device.setScaleY(scale[1]);
        device.setScaleZ(scale[2]);

        // get rotation scale
        SFVec3f rvec = (SFVec3f) getField("rotationScale");
        float[] rscale = new float[3];
        rvec.getValue(rscale);
        device.setScaleR(rscale[0]);
        device.setScaleU(rscale[1]);
        device.setScaleV(rscale[2]);

        // initialize enabled flag
        enabled = ((SFBool) getField("enabled")).getValue();

        // initialize latching flag
        latching = ((SFBool) getField("latching")).getValue();

        // set the description
        SFString description = (SFString) getEventOut("description");
        description.setValue(device.getDescription());

        // get references to the output fields
        position = (SFVec3f) getEventOut("position");
        orientation = (SFRotation) getEventOut("orientation");
        buttons = (SFInt32) getEventOut("buttons");
        whichButton = (SFInt32) getEventOut("whichButton");
        for (int i = 1; i <= 8; ++i)
            buttonArray[i-1] = (SFBool) getEventOut("button" + i);
    }

    public void processEvent(Event ev) {
        if (ev.getName().equals("pollTime")) {
            if (!enabled)  // do nothing if we're not enabled
                return;
            device.poll();
            if (device.getStatus() != 0)  // device error
                return;
            if (device.hasChanged()) {
                position.setValue(device.getX(), device.getY(), device.getZ());
                float r = device.getR(), u = device.getU(), v = device.getV();
                if (r == 0 && u == 0 && v == 0)  // no rotation, common case
                    orientation.setValue(0, 0, 1, 0);
                else {
                    Quaternion q = new Quaternion(u, r, v);
                    orientation.setValue(q.toAxisAndAngle());
                }
            }
            if (device.buttonHasChanged()) {
                int b = device.getButtons();
                if (!latching || device.getButtons() != 0)
                    whichButton.setValue(device.getButton());
                for (int i = 0; i < 8; ++i)
                    buttonArray[i].setValue((b & (1 << i)) != 0);
                buttons.setValue(device.getButtons());
            }
        }
        else if (ev.getName().equals("enabled"))
            enabled = ((ConstSFBool) ev.getValue()).getValue();
        else if (ev.getName().equals("level")) {
            // set all rotations to zero
            if (((ConstSFBool) ev.getValue()).getValue()) {
                device.setR(0);
                device.setU(0);
                device.setV(0);
            }
        }
        else if (ev.getName().equals("home")) {
            // set all translations to zero
            if (((ConstSFBool) ev.getValue()).getValue()) {
                device.setX(0);
                device.setY(0);
                device.setZ(0);
            }
        }
        else if (ev.getName().equals("set_accumulateTranslations")) {
            device.setAccumulateTranslations(((ConstSFBool) ev.getValue()).getValue());
        }
        else if (ev.getName().equals("set_accumulateRotations")) {
            device.setAccumulateRotations(((ConstSFBool) ev.getValue()).getValue());
        }
    }
}

